-------------------------------------------------------------------------------
-- (C) Altran Praxis Limited
-------------------------------------------------------------------------------
--
-- The SPARK toolset is free software; you can redistribute it and/or modify it
-- under terms of the GNU General Public License as published by the Free
-- Software Foundation; either version 3, or (at your option) any later
-- version. The SPARK toolset is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
-- Public License for more details. You should have received a copy of the GNU
-- General Public License distributed with the SPARK toolset; see file
-- COPYING3. If not, go to http://www.gnu.org/licenses for a complete copy of
-- the license.
--
--=============================================================================

separate (Dictionary)
function IsExport
  (Abstraction  : Abstractions;
   TheProcedure : Symbol;
   Variable     : Symbol)
  return         Boolean
is

   --------------------------------------------------------------------------------

   function IsExportedParameter (Abstraction             : Abstractions;
                                 TheProcedure, Parameter : Symbol) return Boolean
   --# global in Dict;
   is
   begin
      return IsFormalParameter (TheProcedure, Parameter) and then RawDict.GetSubprogramParameterExported (Abstraction, Parameter);
   end IsExportedParameter;

   --------------------------------------------------------------------------------

   function IsExportedVariable
     (Abstraction  : Abstractions;
      TheProcedure : Symbol;
      Variable     : Symbol)
     return         Boolean
   --# global in Dict;
   is
      Result         : Boolean;
      GlobalVariable : Symbol;
   begin

      -- allow use with task type as well as procedure
      if IsSubprogram (TheProcedure) then
         GlobalVariable := RawDict.GetSubprogramFirstGlobalVariable (Abstraction, TheProcedure);
      else
         GlobalVariable := RawDict.GetTaskTypeFirstGlobalVariable (Abstraction, TheProcedure);
      end if;

      loop
         if GlobalVariable = NullSymbol then
            Result := False;
            exit;
         end if;
         if RawDict.GetGlobalVariableVariable (GlobalVariable) = Variable then
            Result := RawDict.GetGlobalVariableExported (GlobalVariable);
            exit;
         end if;
         GlobalVariable := RawDict.GetNextGlobalVariable (GlobalVariable);
      end loop;

      return Result;

   end IsExportedVariable;

   --------------------------------------------------------------------------------

begin
   return IsExportedParameter (Abstraction, TheProcedure, Variable)
     or else IsExportedVariable (Abstraction, TheProcedure, Variable);
end IsExport;
