-------------------------------------------------------------------------------
-- (C) Altran Praxis Limited
-------------------------------------------------------------------------------
--
-- The SPARK toolset is free software; you can redistribute it and/or modify it
-- under terms of the GNU General Public License as published by the Free
-- Software Foundation; either version 3, or (at your option) any later
-- version. The SPARK toolset is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
-- Public License for more details. You should have received a copy of the GNU
-- General Public License distributed with the SPARK toolset; see file
-- COPYING3. If not, go to http://www.gnu.org/licenses for a complete copy of
-- the license.
--
--=============================================================================

with SLI;

separate (Sem.CompUnit)
procedure Wf_Subtype_Declaration (Node  : in STree.SyntaxNode;
                                  Scope : in Dictionary.Scopes) is
   type Real_Type is (Is_Floating, Is_Fixed);

   Type_Node, Ident_Node, Constraint_Node : STree.SyntaxNode;
   Id_Str                                 : LexTokenManager.Lex_String;
   Constraint_Found, Ok_To_Add            : Boolean;
   All_Indexes_OK                         : Boolean;
   Subtype_Sym, Type_Sym                  : Dictionary.Symbol;
   Subtype_Is_Static                      : Boolean;
   Lower, Upper, Accuracy                 : LexTokenManager.Lex_String;
   Subtype_Symbol                         : Dictionary.Symbol;

   -------------------------------------------------------------------

   function Is_Named_Association (Node : STree.SyntaxNode) return Boolean
   --# global in STree.Table;
   is
      Current_Node : STree.SyntaxNode;
   begin
      -- ASSUME Node = index_or_discriminant_constraint
      SystemErrors.RT_Assert
        (C       => Syntax_Node_Type (Node => Node) = SPSymbols.index_or_discriminant_constraint,
         Sys_Err => SystemErrors.Invalid_Syntax_Tree,
         Msg     => "Expect Node = index_or_discriminant_constraint in Is_Named_Association");
      Current_Node := Child_Node (Current_Node => Child_Node (Current_Node => Node));
      -- ASSUME Current_Node = named_argument_association OR positional_argument_association
      SystemErrors.RT_Assert
        (C       => Syntax_Node_Type (Node => Current_Node) = SPSymbols.named_argument_association
           or else Syntax_Node_Type (Node => Current_Node) = SPSymbols.positional_argument_association,
         Sys_Err => SystemErrors.Invalid_Syntax_Tree,
         Msg     => "Expect Current_Node = named_argument_association OR positional_argument_association in Is_Named_Association");
      return Syntax_Node_Type (Node => Current_Node) = SPSymbols.named_argument_association;
   end Is_Named_Association;

   -------------------------------------------------------------------

   procedure Check_Index_Constraints
     (Constraint_Node : in     STree.SyntaxNode;
      Scope           : in     Dictionary.Scopes;
      Subtype_Sym     : in     Dictionary.Symbol;
      Type_Sym        : in     Dictionary.Symbol;
      All_Indexes_OK  :    out Boolean)
   --# global in     CommandLineData.Content;
   --#        in     ContextManager.Ops.File_Heap;
   --#        in     ContextManager.Ops.Unit_Heap;
   --#        in     ContextManager.Ops.Unit_Stack;
   --#        in out Dictionary.Dict;
   --#        in out ErrorHandler.Error_Context;
   --#        in out LexTokenManager.State;
   --#        in out SLI.State;
   --#        in out SPARK_IO.File_Sys;
   --#        in out Statistics.TableUsage;
   --#        in out STree.Table;
   --#           out AggregateStack.State;
   --#           out TheHeap;
   --# derives AggregateStack.State,
   --#         All_Indexes_OK,
   --#         Dictionary.Dict,
   --#         LexTokenManager.State,
   --#         STree.Table,
   --#         TheHeap                    from CommandLineData.Content,
   --#                                         Constraint_Node,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         LexTokenManager.State,
   --#                                         Scope,
   --#                                         STree.Table,
   --#                                         Subtype_Sym,
   --#                                         Type_Sym &
   --#         ErrorHandler.Error_Context,
   --#         SLI.State,
   --#         SPARK_IO.File_Sys          from CommandLineData.Content,
   --#                                         Constraint_Node,
   --#                                         ContextManager.Ops.File_Heap,
   --#                                         ContextManager.Ops.Unit_Heap,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         ErrorHandler.Error_Context,
   --#                                         LexTokenManager.State,
   --#                                         Scope,
   --#                                         SLI.State,
   --#                                         SPARK_IO.File_Sys,
   --#                                         STree.Table,
   --#                                         Subtype_Sym,
   --#                                         Type_Sym &
   --#         Statistics.TableUsage      from *,
   --#                                         CommandLineData.Content,
   --#                                         Constraint_Node,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         LexTokenManager.State,
   --#                                         Scope,
   --#                                         STree.Table,
   --#                                         Subtype_Sym,
   --#                                         Type_Sym;
   is
      Index_Iterator                              : Dictionary.Iterator;
      Current_Constraint_Node                     : STree.SyntaxNode;
      Index_Sym, Constraint_Sym                   : Dictionary.Symbol;
      Indexes_Finished, Constraints_Finished      : Boolean;
      Result_Of_First_Check, Result_Of_Last_Check : Maths.Value;
      Constraint_First, Constraint_Last           : LexTokenManager.Lex_String;

      procedure Get_First_Index
        (Type_Sym         : in     Dictionary.Symbol;
         Index_Sym        :    out Dictionary.Symbol;
         Index_Iterator   :    out Dictionary.Iterator;
         Indexes_Finished :    out Boolean)
      --# global in Dictionary.Dict;
      --# derives Indexes_Finished,
      --#         Index_Iterator,
      --#         Index_Sym        from Dictionary.Dict,
      --#                               Type_Sym;
      is
      begin
         Index_Iterator := Dictionary.FirstArrayIndex (Type_Sym);
         if Dictionary.IsNullIterator (Index_Iterator) then
            Index_Sym        := Dictionary.NullSymbol;
            Indexes_Finished := True;
         else
            Index_Sym        := Dictionary.CurrentSymbol (Index_Iterator);
            Indexes_Finished := False;
         end if;
      end Get_First_Index;

      procedure Get_Next_Index
        (Index_Sym        :    out Dictionary.Symbol;
         Index_Iterator   : in out Dictionary.Iterator;
         Indexes_Finished :    out Boolean)
      --# global in Dictionary.Dict;
      --# derives Indexes_Finished,
      --#         Index_Iterator,
      --#         Index_Sym        from Dictionary.Dict,
      --#                               Index_Iterator;
      is
      begin
         Index_Iterator := Dictionary.NextSymbol (Index_Iterator);
         if Dictionary.IsNullIterator (Index_Iterator) then
            Index_Sym        := Dictionary.NullSymbol;
            Indexes_Finished := True;
         else
            Index_Sym        := Dictionary.CurrentSymbol (Index_Iterator);
            Indexes_Finished := False;
         end if;
      end Get_Next_Index;

      procedure Walk_Expression_As_Type_Mark
        (Exp_Node       : in     STree.SyntaxNode;
         Scope          : in     Dictionary.Scopes;
         Constraint_Sym :    out Dictionary.Symbol)
      --# global in     CommandLineData.Content;
      --#        in     ContextManager.Ops.File_Heap;
      --#        in     ContextManager.Ops.Unit_Heap;
      --#        in     ContextManager.Ops.Unit_Stack;
      --#        in out Dictionary.Dict;
      --#        in out ErrorHandler.Error_Context;
      --#        in out LexTokenManager.State;
      --#        in out SLI.State;
      --#        in out SPARK_IO.File_Sys;
      --#        in out Statistics.TableUsage;
      --#        in out STree.Table;
      --#           out AggregateStack.State;
      --#           out TheHeap;
      --# derives AggregateStack.State,
      --#         Constraint_Sym,
      --#         Dictionary.Dict,
      --#         LexTokenManager.State,
      --#         STree.Table,
      --#         TheHeap                    from CommandLineData.Content,
      --#                                         ContextManager.Ops.Unit_Stack,
      --#                                         Dictionary.Dict,
      --#                                         Exp_Node,
      --#                                         LexTokenManager.State,
      --#                                         Scope,
      --#                                         STree.Table &
      --#         ErrorHandler.Error_Context,
      --#         SLI.State,
      --#         SPARK_IO.File_Sys          from CommandLineData.Content,
      --#                                         ContextManager.Ops.File_Heap,
      --#                                         ContextManager.Ops.Unit_Heap,
      --#                                         ContextManager.Ops.Unit_Stack,
      --#                                         Dictionary.Dict,
      --#                                         ErrorHandler.Error_Context,
      --#                                         Exp_Node,
      --#                                         LexTokenManager.State,
      --#                                         Scope,
      --#                                         SLI.State,
      --#                                         SPARK_IO.File_Sys,
      --#                                         STree.Table &
      --#         Statistics.TableUsage      from *,
      --#                                         CommandLineData.Content,
      --#                                         ContextManager.Ops.Unit_Stack,
      --#                                         Dictionary.Dict,
      --#                                         Exp_Node,
      --#                                         LexTokenManager.State,
      --#                                         Scope,
      --#                                         STree.Table;
      is
         Unwanted_Seq          : SeqAlgebra.Seq;
         Unused_Component_Data : ComponentManager.ComponentData;
         Constraint_Record     : Exp_Record;
      begin
         -- ASSUME Exp_Node = expression
         SystemErrors.RT_Assert
           (C       => Syntax_Node_Type (Node => Exp_Node) = SPSymbols.expression,
            Sys_Err => SystemErrors.Invalid_Syntax_Tree,
            Msg     => "Expect Exp_Node = expression in Walk_Expression_As_Type_Mark");

         Heap.Initialize (TheHeap);
         SeqAlgebra.CreateSeq (TheHeap, Unwanted_Seq);
         ComponentManager.Initialise (Unused_Component_Data);
         --# accept Flow, 10, Unused_Component_Data, "Expected ineffective assignment";
         WalkExpression
           (Exp_Node                => Exp_Node,
            Scope                   => Scope,
            Type_Context            => Dictionary.GetUnknownTypeMark,
            Context_Requires_Static => False,
            Result                  => Constraint_Record,
            Ref_Var                 => Unwanted_Seq,
            Component_Data          => Unused_Component_Data);
         --# end accept;
         SeqAlgebra.DisposeOfSeq (TheHeap, Unwanted_Seq);

         if Constraint_Record.Is_ARange then
            Constraint_Sym := Constraint_Record.Type_Symbol;
         else
            ErrorHandler.Semantic_Error
              (Err_Num   => 95,
               Reference => ErrorHandler.No_Reference,
               Position  => Node_Position (Node => Exp_Node),
               Id_Str    => LexTokenManager.Null_String);
            Constraint_Sym := Dictionary.GetUnknownTypeMark;
         end if;
         Heap.ReportUsage (TheHeap);
      end Walk_Expression_As_Type_Mark;

      procedure Get_First_Constraint
        (Constraint_Node         : in     STree.SyntaxNode;
         Scope                   : in     Dictionary.Scopes;
         Current_Constraint_Node :    out STree.SyntaxNode;
         Constraint_Sym          :    out Dictionary.Symbol;
         Constraints_Finished    :    out Boolean)
      --# global in     CommandLineData.Content;
      --#        in     ContextManager.Ops.File_Heap;
      --#        in     ContextManager.Ops.Unit_Heap;
      --#        in     ContextManager.Ops.Unit_Stack;
      --#        in out Dictionary.Dict;
      --#        in out ErrorHandler.Error_Context;
      --#        in out LexTokenManager.State;
      --#        in out SLI.State;
      --#        in out SPARK_IO.File_Sys;
      --#        in out Statistics.TableUsage;
      --#        in out STree.Table;
      --#           out AggregateStack.State;
      --#           out TheHeap;
      --# derives AggregateStack.State,
      --#         Constraint_Sym,
      --#         Dictionary.Dict,
      --#         LexTokenManager.State,
      --#         STree.Table,
      --#         TheHeap                    from CommandLineData.Content,
      --#                                         Constraint_Node,
      --#                                         ContextManager.Ops.Unit_Stack,
      --#                                         Dictionary.Dict,
      --#                                         LexTokenManager.State,
      --#                                         Scope,
      --#                                         STree.Table &
      --#         Constraints_Finished       from  &
      --#         Current_Constraint_Node    from Constraint_Node,
      --#                                         STree.Table &
      --#         ErrorHandler.Error_Context,
      --#         SLI.State,
      --#         SPARK_IO.File_Sys          from CommandLineData.Content,
      --#                                         Constraint_Node,
      --#                                         ContextManager.Ops.File_Heap,
      --#                                         ContextManager.Ops.Unit_Heap,
      --#                                         ContextManager.Ops.Unit_Stack,
      --#                                         Dictionary.Dict,
      --#                                         ErrorHandler.Error_Context,
      --#                                         LexTokenManager.State,
      --#                                         Scope,
      --#                                         SLI.State,
      --#                                         SPARK_IO.File_Sys,
      --#                                         STree.Table &
      --#         Statistics.TableUsage      from *,
      --#                                         CommandLineData.Content,
      --#                                         Constraint_Node,
      --#                                         ContextManager.Ops.Unit_Stack,
      --#                                         Dictionary.Dict,
      --#                                         LexTokenManager.State,
      --#                                         Scope,
      --#                                         STree.Table;
      is
      begin
         -- ASSUME Constraint_Node = index_or_discriminant_constraint
         SystemErrors.RT_Assert
           (C       => Syntax_Node_Type (Node => Constraint_Node) = SPSymbols.index_or_discriminant_constraint,
            Sys_Err => SystemErrors.Invalid_Syntax_Tree,
            Msg     => "Expect Constraint_Node = index_or_discriminant_constraint in Get_First_Constraint");

         Constraints_Finished    := False;
         Current_Constraint_Node := Child_Node (Current_Node => Child_Node (Current_Node => Constraint_Node));
         while Syntax_Node_Type (Node => Current_Constraint_Node) /= SPSymbols.expression loop
            -- ASSUME Current_Constraint_Node = positional_argument_association
            SystemErrors.RT_Assert
              (C       => Syntax_Node_Type (Node => Current_Constraint_Node) = SPSymbols.positional_argument_association,
               Sys_Err => SystemErrors.Invalid_Syntax_Tree,
               Msg     => "Expect Current_Constraint_Node = positional_argument_association in Get_First_Constraint");
            Current_Constraint_Node := Child_Node (Current_Node => Current_Constraint_Node);
         end loop;
         -- ASSUME Current_Constraint_Node = expression
         SystemErrors.RT_Assert
           (C       => Syntax_Node_Type (Node => Current_Constraint_Node) = SPSymbols.expression,
            Sys_Err => SystemErrors.Invalid_Syntax_Tree,
            Msg     => "Expect Current_Constraint_Node = expression in Get_First_Constraint");
         Walk_Expression_As_Type_Mark (Exp_Node       => Current_Constraint_Node,
                                       Scope          => Scope,
                                       Constraint_Sym => Constraint_Sym);
      end Get_First_Constraint;

      procedure Get_Next_Constraint
        (Scope                   : in     Dictionary.Scopes;
         Current_Constraint_Node : in out STree.SyntaxNode;
         Constraint_Sym          :    out Dictionary.Symbol;
         Constraints_Finished    :    out Boolean)
      --# global in     CommandLineData.Content;
      --#        in     ContextManager.Ops.File_Heap;
      --#        in     ContextManager.Ops.Unit_Heap;
      --#        in     ContextManager.Ops.Unit_Stack;
      --#        in out AggregateStack.State;
      --#        in out Dictionary.Dict;
      --#        in out ErrorHandler.Error_Context;
      --#        in out LexTokenManager.State;
      --#        in out SLI.State;
      --#        in out SPARK_IO.File_Sys;
      --#        in out Statistics.TableUsage;
      --#        in out STree.Table;
      --#        in out TheHeap;
      --# derives AggregateStack.State,
      --#         Dictionary.Dict,
      --#         LexTokenManager.State,
      --#         Statistics.TableUsage,
      --#         STree.Table,
      --#         TheHeap                    from *,
      --#                                         CommandLineData.Content,
      --#                                         ContextManager.Ops.Unit_Stack,
      --#                                         Current_Constraint_Node,
      --#                                         Dictionary.Dict,
      --#                                         LexTokenManager.State,
      --#                                         Scope,
      --#                                         STree.Table &
      --#         Constraints_Finished,
      --#         Current_Constraint_Node    from Current_Constraint_Node,
      --#                                         STree.Table &
      --#         Constraint_Sym             from CommandLineData.Content,
      --#                                         ContextManager.Ops.Unit_Stack,
      --#                                         Current_Constraint_Node,
      --#                                         Dictionary.Dict,
      --#                                         LexTokenManager.State,
      --#                                         Scope,
      --#                                         STree.Table &
      --#         ErrorHandler.Error_Context,
      --#         SLI.State,
      --#         SPARK_IO.File_Sys          from CommandLineData.Content,
      --#                                         ContextManager.Ops.File_Heap,
      --#                                         ContextManager.Ops.Unit_Heap,
      --#                                         ContextManager.Ops.Unit_Stack,
      --#                                         Current_Constraint_Node,
      --#                                         Dictionary.Dict,
      --#                                         ErrorHandler.Error_Context,
      --#                                         LexTokenManager.State,
      --#                                         Scope,
      --#                                         SLI.State,
      --#                                         SPARK_IO.File_Sys,
      --#                                         STree.Table;
      is
      begin
         -- ASSUME Current_Constraint_Node = expression
         SystemErrors.RT_Assert
           (C       => Syntax_Node_Type (Node => Current_Constraint_Node) = SPSymbols.expression,
            Sys_Err => SystemErrors.Invalid_Syntax_Tree,
            Msg     => "Expect Current_Constraint_Node = expression in Get_Next_Constraint");
         Current_Constraint_Node := Next_Sibling (Current_Node => Parent_Node (Current_Node => Current_Constraint_Node));
         -- ASSUME Current_Constraint_Node = expression OR NULL
         if Current_Constraint_Node = STree.NullNode then
            -- ASSUME Current_Constraint_Node = NULL
            Constraints_Finished := True;
            Constraint_Sym       := Dictionary.NullSymbol;
         elsif Syntax_Node_Type (Node => Current_Constraint_Node) = SPSymbols.expression then
            -- ASSUME Current_Constraint_Node = expression
            Constraints_Finished := False;
            Walk_Expression_As_Type_Mark (Exp_Node       => Current_Constraint_Node,
                                          Scope          => Scope,
                                          Constraint_Sym => Constraint_Sym);
         else
            Constraints_Finished := True;
            Constraint_Sym       := Dictionary.NullSymbol;
            SystemErrors.Fatal_Error
              (Sys_Err => SystemErrors.Invalid_Syntax_Tree,
               Msg     => "Expect Current_Constraint_Node = expression OR NULL in Get_Next_Constraint");
         end if;
      end Get_Next_Constraint;

   begin -- Check_Index_Constraints

      -- ASSUME Constraint_Node = index_or_discriminant_constraint
      SystemErrors.RT_Assert
        (C       => Syntax_Node_Type (Node => Constraint_Node) = SPSymbols.index_or_discriminant_constraint,
         Sys_Err => SystemErrors.Invalid_Syntax_Tree,
         Msg     => "Expect Constraint_Node = index_or_discriminant_constraint in Check_Index_Constraints");

      All_Indexes_OK := True;
      Get_First_Index
        (Type_Sym         => Type_Sym,
         Index_Sym        => Index_Sym,
         Index_Iterator   => Index_Iterator,
         Indexes_Finished => Indexes_Finished);
      Get_First_Constraint
        (Constraint_Node         => Constraint_Node,
         Scope                   => Scope,
         Current_Constraint_Node => Current_Constraint_Node,
         Constraint_Sym          => Constraint_Sym,
         Constraints_Finished    => Constraints_Finished);
      while not (Indexes_Finished or else Constraints_Finished) loop
         -- ASSUME Current_Constraint_Node = expression OR NULL
         SystemErrors.RT_Assert
           (C       => Current_Constraint_Node = STree.NullNode
              or else Syntax_Node_Type (Node => Current_Constraint_Node) = SPSymbols.expression,
            Sys_Err => SystemErrors.Invalid_Syntax_Tree,
            Msg     => "Expect Current_Constraint_Node = expression OR NULL in Check_Index_Constraints");
         if not Dictionary.CompatibleTypes (Scope, Index_Sym, Constraint_Sym) then
            ErrorHandler.Semantic_Error_Sym2
              (Err_Num   => 107,
               Reference => ErrorHandler.No_Reference,
               Position  => Node_Position (Node => Current_Constraint_Node),
               Sym       => Constraint_Sym,
               Sym2      => Index_Sym,
               Scope     => Scope);
            All_Indexes_OK := False;
         end if;

         Constraint_First := Dictionary.GetScalarAttributeValue (False, LexTokenManager.First_Token, Constraint_Sym);
         Constraint_Last  := Dictionary.GetScalarAttributeValue (False, LexTokenManager.Last_Token, Constraint_Sym);

         -- Check that Constraint_Sym'First is OK wrt Index_Sym'First
         if Dictionary.IsPredefinedStringType (Type_Sym) then
            if LexTokenManager.Lex_String_Case_Insensitive_Compare
              (Lex_Str1 => Constraint_First,
               Lex_Str2 => LexTokenManager.One_Value) /=
              LexTokenManager.Str_Eq then
               ErrorHandler.Semantic_Error
                 (Err_Num   => 417,
                  Reference => ErrorHandler.No_Reference,
                  Position  => Node_Position (Node => Constraint_Node),
                  Id_Str    => LexTokenManager.Null_String);
            end if;
         else
            ConstraintCheck
              (Maths.ValueRep (Constraint_First),
               Result_Of_First_Check,
               False, -- cannot be in annotation here
               Index_Sym,
               Node_Position (Node => Current_Constraint_Node));
            if Result_Of_First_Check = Maths.NoValue then
               All_Indexes_OK := False;
            end if;

         end if;

         -- Check that Constraint_Sym'Last is OK wrt Index_Sym'Last
         ConstraintCheck
           (Maths.ValueRep (Constraint_Last),
            Result_Of_Last_Check,
            False, -- cannot be in annotation here
            Index_Sym,
            Node_Position (Node => Current_Constraint_Node));
         if Result_Of_Last_Check = Maths.NoValue then
            All_Indexes_OK := False;
         end if;

         Dictionary.AddArrayIndex
           (TheArrayType       => Subtype_Sym,
            TheIndexType       => Constraint_Sym,
            Comp_Unit          => ContextManager.Ops.Current_Unit,
            IndexTypeReference => Dictionary.Location'(Start_Position => Node_Position (Node => Current_Constraint_Node),
                                                       End_Position   => Node_Position (Node => Current_Constraint_Node)));
         Get_Next_Index (Index_Sym        => Index_Sym,
                         Index_Iterator   => Index_Iterator,
                         Indexes_Finished => Indexes_Finished);
         Get_Next_Constraint
           (Scope                   => Scope,
            Current_Constraint_Node => Current_Constraint_Node,
            Constraint_Sym          => Constraint_Sym,
            Constraints_Finished    => Constraints_Finished);
      end loop;
      if not (Indexes_Finished and then Constraints_Finished) then
         ErrorHandler.Semantic_Error
           (Err_Num   => 93,
            Reference => ErrorHandler.No_Reference,
            Position  => Node_Position (Node => Constraint_Node),
            Id_Str    => LexTokenManager.Null_String);
         All_Indexes_OK := False;
      end if;
   end Check_Index_Constraints;

   -----------------------------------------------------------------------

   procedure Check_Real_Accuracy
     (Node     : in     STree.SyntaxNode;
      Scope    : in     Dictionary.Scopes;
      Sort     : in     Real_Type;
      Accuracy :    out LexTokenManager.Lex_String;
      Static   :    out Boolean)
   --# global in     CommandLineData.Content;
   --#        in     ContextManager.Ops.File_Heap;
   --#        in     ContextManager.Ops.Unit_Heap;
   --#        in     ContextManager.Ops.Unit_Stack;
   --#        in out AggregateStack.State;
   --#        in out Dictionary.Dict;
   --#        in out ErrorHandler.Error_Context;
   --#        in out LexTokenManager.State;
   --#        in out SLI.State;
   --#        in out SPARK_IO.File_Sys;
   --#        in out Statistics.TableUsage;
   --#        in out STree.Table;
   --#           out TheHeap;
   --# derives Accuracy                   from CommandLineData.Content,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         LexTokenManager.State,
   --#                                         Node,
   --#                                         Scope,
   --#                                         Sort,
   --#                                         STree.Table &
   --#         AggregateStack.State,
   --#         Dictionary.Dict,
   --#         LexTokenManager.State,
   --#         Statistics.TableUsage,
   --#         STree.Table                from *,
   --#                                         CommandLineData.Content,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         LexTokenManager.State,
   --#                                         Node,
   --#                                         Scope,
   --#                                         STree.Table &
   --#         ErrorHandler.Error_Context,
   --#         SPARK_IO.File_Sys          from CommandLineData.Content,
   --#                                         ContextManager.Ops.File_Heap,
   --#                                         ContextManager.Ops.Unit_Heap,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         ErrorHandler.Error_Context,
   --#                                         LexTokenManager.State,
   --#                                         Node,
   --#                                         Scope,
   --#                                         SLI.State,
   --#                                         Sort,
   --#                                         SPARK_IO.File_Sys,
   --#                                         STree.Table &
   --#         SLI.State                  from *,
   --#                                         CommandLineData.Content,
   --#                                         ContextManager.Ops.File_Heap,
   --#                                         ContextManager.Ops.Unit_Heap,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         ErrorHandler.Error_Context,
   --#                                         LexTokenManager.State,
   --#                                         Node,
   --#                                         Scope,
   --#                                         SPARK_IO.File_Sys,
   --#                                         STree.Table &
   --#         Static,
   --#         TheHeap                    from CommandLineData.Content,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         LexTokenManager.State,
   --#                                         Node,
   --#                                         Scope,
   --#                                         STree.Table;
   is
      Exp_Node              : STree.SyntaxNode;
      Accuracy_Type         : Exp_Record;
      Unwanted_Seq          : SeqAlgebra.Seq;
      Unused_Component_Data : ComponentManager.ComponentData;

      -------------------------------

      function Type_Correct
        (Type_Sym : Dictionary.Symbol;
         Scope    : Dictionary.Scopes;
         Sort     : Real_Type)
        return     Boolean
      --# global in Dictionary.Dict;
      is
         Result : Boolean;
      begin
         if Sort = Is_Floating then
            Result := Dictionary.IsIntegerTypeMark (Type_Sym, Scope);
         else
            Result := Dictionary.IsRealTypeMark (Type_Sym, Scope);
         end if;
         return Result or else Dictionary.IsUnknownTypeMark (Type_Sym);
      end Type_Correct;

   begin -- Check_Real_Accuracy

      -- ASSUME Node = floating_point_constraint OR fixed_point_constraint
      SystemErrors.RT_Assert
        (C       => Syntax_Node_Type (Node => Node) = SPSymbols.floating_point_constraint
           or else Syntax_Node_Type (Node => Node) = SPSymbols.fixed_point_constraint,
         Sys_Err => SystemErrors.Invalid_Syntax_Tree,
         Msg     => "Expect Node = floating_point_constraint OR fixed_point_constraint in Check_Real_Accuracy");

      Heap.Initialize (TheHeap);
      ComponentManager.Initialise (Unused_Component_Data);

      case CommandLineData.Content.Language_Profile is
         when CommandLineData.SPARK83 =>
            Exp_Node := Child_Node (Current_Node => Child_Node (Current_Node => Node));
            -- ASSUME Exp_Node = simple_expression
            SystemErrors.RT_Assert
              (C       => Syntax_Node_Type (Node => Exp_Node) = SPSymbols.simple_expression,
               Sys_Err => SystemErrors.Invalid_Syntax_Tree,
               Msg     => "Expect Exp_Node = simple_expression in Check_Real_Accuracy");
            SeqAlgebra.CreateSeq (TheHeap, Unwanted_Seq);
            --# accept Flow, 10, Unused_Component_Data, "Expected ineffective assignment";
            WalkExpression
              (Exp_Node                => Exp_Node,
               Scope                   => Scope,
               Type_Context            => Dictionary.GetUnknownTypeMark,
               Context_Requires_Static => True,
               Result                  => Accuracy_Type,
               Ref_Var                 => Unwanted_Seq,
               Component_Data          => Unused_Component_Data);
            --# end accept;
            SeqAlgebra.DisposeOfSeq (TheHeap, Unwanted_Seq);
            Maths.StorageRep (Accuracy_Type.Value, Accuracy);
            if not Type_Correct (Type_Sym => Accuracy_Type.Type_Symbol,
                                 Scope    => Scope,
                                 Sort     => Sort) then
               Accuracy := LexTokenManager.Null_String;
               ErrorHandler.Semantic_Error
                 (Err_Num   => 38,
                  Reference => ErrorHandler.No_Reference,
                  Position  => Node_Position (Node => Node),
                  Id_Str    => LexTokenManager.Null_String);
            end if;
            Static := Accuracy_Type.Is_Static;
         when CommandLineData.SPARK95 | CommandLineData.SPARK2005 =>
            -- reduced accuracy subtypes of reals are not allowed in 95 or 2005
            Static   := True; -- to reduce knock-on errors
            Accuracy := LexTokenManager.Null_String;
            ErrorHandler.Semantic_Error
              (Err_Num   => 608,
               Reference => 9,
               Position  => Node_Position (Node => Node),
               Id_Str    => LexTokenManager.Null_String);
      end case;
      Heap.ReportUsage (TheHeap);
   end Check_Real_Accuracy;

   ------------------------------------------------------------------------

   procedure Check_Range
     (Node         : in     STree.SyntaxNode;
      Type_Sym     : in     Dictionary.Symbol;
      Scope        : in     Dictionary.Scopes;
      Static       : in out Boolean;
      Lower, Upper :    out LexTokenManager.Lex_String)
   --# global in     CommandLineData.Content;
   --#        in     ContextManager.Ops.File_Heap;
   --#        in     ContextManager.Ops.Unit_Heap;
   --#        in     ContextManager.Ops.Unit_Stack;
   --#        in out Dictionary.Dict;
   --#        in out ErrorHandler.Error_Context;
   --#        in out LexTokenManager.State;
   --#        in out SLI.State;
   --#        in out SPARK_IO.File_Sys;
   --#        in out Statistics.TableUsage;
   --#        in out STree.Table;
   --#           out AggregateStack.State;
   --#           out TheHeap;
   --# derives AggregateStack.State,
   --#         Dictionary.Dict,
   --#         LexTokenManager.State,
   --#         Lower,
   --#         STree.Table,
   --#         TheHeap,
   --#         Upper                      from CommandLineData.Content,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         LexTokenManager.State,
   --#                                         Node,
   --#                                         Scope,
   --#                                         STree.Table,
   --#                                         Type_Sym &
   --#         ErrorHandler.Error_Context,
   --#         SLI.State,
   --#         SPARK_IO.File_Sys          from CommandLineData.Content,
   --#                                         ContextManager.Ops.File_Heap,
   --#                                         ContextManager.Ops.Unit_Heap,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         ErrorHandler.Error_Context,
   --#                                         LexTokenManager.State,
   --#                                         Node,
   --#                                         Scope,
   --#                                         SLI.State,
   --#                                         SPARK_IO.File_Sys,
   --#                                         STree.Table,
   --#                                         Type_Sym &
   --#         Static,
   --#         Statistics.TableUsage      from *,
   --#                                         CommandLineData.Content,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         LexTokenManager.State,
   --#                                         Node,
   --#                                         Scope,
   --#                                         STree.Table,
   --#                                         Type_Sym;
   is
      Range_Result                                               : Exp_Record;
      Unwanted_Seq                                               : SeqAlgebra.Seq;
      Unused_Component_Data                                      : ComponentManager.ComponentData;
      Upper_After_Constraint_Check, Lower_After_Constraint_Check : Maths.Value;
      RHS_Node                                                   : STree.SyntaxNode;
   begin
      -- ASSUME Node = arange
      SystemErrors.RT_Assert
        (C       => Syntax_Node_Type (Node => Node) = SPSymbols.arange,
         Sys_Err => SystemErrors.Invalid_Syntax_Tree,
         Msg     => "Expect Node = arange in Check_Range");

      Heap.Initialize (TheHeap);
      SeqAlgebra.CreateSeq (TheHeap, Unwanted_Seq);
      ComponentManager.Initialise (Unused_Component_Data);
      --# accept Flow, 10, Unused_Component_Data, "Expected ineffective assignment";
      WalkExpression
        (Exp_Node                => Node,
         Scope                   => Scope,
         Type_Context            => Type_Sym,
         Context_Requires_Static => False,
         Result                  => Range_Result,
         Ref_Var                 => Unwanted_Seq,
         Component_Data          => Unused_Component_Data);
      --# end accept;
      SeqAlgebra.DisposeOfSeq (TheHeap, Unwanted_Seq);
      Static := Static and then Range_Result.Is_Static;
      Maths.StorageRep (Range_Result.Value, Lower);
      Maths.StorageRep (Range_Result.Range_RHS, Upper);

      -- check that range is constant
      if not Range_Result.Is_Constant then
         Lower := LexTokenManager.Null_String;
         Upper := LexTokenManager.Null_String;
         ErrorHandler.Semantic_Error
           (Err_Num   => 43,
            Reference => ErrorHandler.No_Reference,
            Position  => Node_Position (Node => Node),
            Id_Str    => LexTokenManager.Null_String);
      end if;
      if not Dictionary.CompatibleTypes (Scope, Range_Result.Type_Symbol, Type_Sym) then
         Lower := LexTokenManager.Null_String;
         Upper := LexTokenManager.Null_String;
         ErrorHandler.Semantic_Error_Sym2
           (Err_Num   => 107,
            Reference => ErrorHandler.No_Reference,
            Position  => Node_Position (Node => Node),
            Sym       => Range_Result.Type_Symbol,
            Sym2      => Type_Sym,
            Scope     => Scope);
      end if;

      -- checks for bounds outside type being constrained
      -- see whether node is attribute or X..Y form and select suitable place
      -- to report errors on upper range bound
      RHS_Node := Child_Node (Current_Node => Node);
      -- ASSUME RHS_Node = attribute OR simple_expression
      if Syntax_Node_Type (Node => RHS_Node) = SPSymbols.attribute then
         -- ASSUME RHS_Node = attribute
         RHS_Node := Node;
      elsif Syntax_Node_Type (Node => RHS_Node) = SPSymbols.simple_expression then
         -- ASSUME RHS_Node = simple_expression
         -- must be of form X..Y
         RHS_Node := Next_Sibling (Current_Node => RHS_Node);
      else
         SystemErrors.RT_Assert
           (C       => Syntax_Node_Type (Node => RHS_Node) = SPSymbols.attribute
              or else Syntax_Node_Type (Node => RHS_Node) = SPSymbols.simple_expression,
            Sys_Err => SystemErrors.Invalid_Syntax_Tree,
            Msg     => "Expect RHS_Node = attribute OR simple_expression in Check_Range");
      end if;
      -- ASSUME RHS_Node = arange OR simple_expression
      SystemErrors.RT_Assert
        (C       => Syntax_Node_Type (Node => RHS_Node) = SPSymbols.arange
           or else Syntax_Node_Type (Node => RHS_Node) = SPSymbols.simple_expression,
         Sys_Err => SystemErrors.Invalid_Syntax_Tree,
         Msg     => "Expect RHS_Node = arange OR simple_expression in Check_Range");

      ConstraintCheck (Maths.ValueRep (Lower), Lower_After_Constraint_Check, False, -- can't be in annotation here
                       Type_Sym, Node_Position (Node => Node));
      Maths.StorageRep (Lower_After_Constraint_Check, Lower);

      ConstraintCheck (Maths.ValueRep (Upper), Upper_After_Constraint_Check, False, -- can't be in annotation here
                       Type_Sym, Node_Position (Node => RHS_Node));
      Maths.StorageRep (Upper_After_Constraint_Check, Upper);

      Heap.ReportUsage (TheHeap);
   end Check_Range;

   -----------------------------------------------------------------------

   procedure Check_Real_Range
     (Node         : in     STree.SyntaxNode;
      Type_Sym     : in     Dictionary.Symbol;
      Scope        : in     Dictionary.Scopes;
      Static       : in out Boolean;
      Lower, Upper :    out LexTokenManager.Lex_String)
   --# global in     CommandLineData.Content;
   --#        in     ContextManager.Ops.File_Heap;
   --#        in     ContextManager.Ops.Unit_Heap;
   --#        in     ContextManager.Ops.Unit_Stack;
   --#        in out AggregateStack.State;
   --#        in out Dictionary.Dict;
   --#        in out ErrorHandler.Error_Context;
   --#        in out LexTokenManager.State;
   --#        in out SLI.State;
   --#        in out SPARK_IO.File_Sys;
   --#        in out Statistics.TableUsage;
   --#        in out STree.Table;
   --#        in out TheHeap;
   --# derives AggregateStack.State,
   --#         Dictionary.Dict,
   --#         LexTokenManager.State,
   --#         Static,
   --#         Statistics.TableUsage,
   --#         STree.Table,
   --#         TheHeap                    from *,
   --#                                         CommandLineData.Content,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         LexTokenManager.State,
   --#                                         Node,
   --#                                         Scope,
   --#                                         STree.Table,
   --#                                         Type_Sym &
   --#         ErrorHandler.Error_Context,
   --#         SLI.State,
   --#         SPARK_IO.File_Sys          from CommandLineData.Content,
   --#                                         ContextManager.Ops.File_Heap,
   --#                                         ContextManager.Ops.Unit_Heap,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         ErrorHandler.Error_Context,
   --#                                         LexTokenManager.State,
   --#                                         Node,
   --#                                         Scope,
   --#                                         SLI.State,
   --#                                         SPARK_IO.File_Sys,
   --#                                         STree.Table,
   --#                                         Type_Sym &
   --#         Lower,
   --#         Upper                      from CommandLineData.Content,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         LexTokenManager.State,
   --#                                         Node,
   --#                                         Scope,
   --#                                         STree.Table,
   --#                                         Type_Sym;
   is
      Range_Node : STree.SyntaxNode;
   begin
      -- ASSUME Node = floating_point_constraint OR fixed_point_constraint
      SystemErrors.RT_Assert
        (C       => Syntax_Node_Type (Node => Node) = SPSymbols.floating_point_constraint
           or else Syntax_Node_Type (Node => Node) = SPSymbols.fixed_point_constraint,
         Sys_Err => SystemErrors.Invalid_Syntax_Tree,
         Msg     => "Expect Node = floating_point_constraint OR fixed_point_constraint in Check_Real_Range");

      Range_Node := Next_Sibling (Current_Node => Child_Node (Current_Node => Node));
      -- ASSUME Range_Node = range_constraint OR NULL
      if Range_Node = STree.NullNode then
         -- ASSUME Range_Node = NULL
         -- no range supplied so range is unchanged from parent type
         Upper :=
           Dictionary.GetScalarAttributeValue (False, --no 'base
                                               LexTokenManager.Last_Token, Type_Sym);
         Lower :=
           Dictionary.GetScalarAttributeValue (False, --no 'base
                                               LexTokenManager.First_Token, Type_Sym);
      elsif Syntax_Node_Type (Node => Range_Node) = SPSymbols.range_constraint then
         -- ASSUME Range_Node = range_constraint
         -- a range is supplied
         Check_Range
           (Node     => Child_Node (Current_Node => Range_Node),
            Type_Sym => Type_Sym,
            Scope    => Scope,
            Static   => Static,
            Lower    => Lower,
            Upper    => Upper);
      else
         Lower := LexTokenManager.Null_String;
         Upper := LexTokenManager.Null_String;
         SystemErrors.Fatal_Error
           (Sys_Err => SystemErrors.Invalid_Syntax_Tree,
            Msg     => "Expect Range_Node = range_constraint OR NULL in Check_Real_Range");
      end if;
   end Check_Real_Range;

   -----------------------------------------------------------------------

   procedure Add_Scalar_Subtype
     (Accuracy          : in LexTokenManager.Lex_String;
      Id_Str            : in LexTokenManager.Lex_String;
      Ident_Node        : in STree.SyntaxNode;
      Type_Node_Pos     : in LexTokenManager.Token_Position;
      Type_Sym          : in Dictionary.Symbol;
      Scope             : in Dictionary.Scopes;
      Subtype_Is_Static : in Boolean;
      Lower, Upper      : in LexTokenManager.Lex_String)
   --# global in     CommandLineData.Content;
   --#        in     ContextManager.Ops.File_Heap;
   --#        in     ContextManager.Ops.Unit_Heap;
   --#        in     ContextManager.Ops.Unit_Stack;
   --#        in     ErrorHandler.Error_Context;
   --#        in     LexTokenManager.State;
   --#        in     STree.Table;
   --#        in out Dictionary.Dict;
   --#        in out SLI.State;
   --#        in out SPARK_IO.File_Sys;
   --# derives Dictionary.Dict   from *,
   --#                                Accuracy,
   --#                                ContextManager.Ops.Unit_Stack,
   --#                                Ident_Node,
   --#                                Id_Str,
   --#                                LexTokenManager.State,
   --#                                Lower,
   --#                                Scope,
   --#                                STree.Table,
   --#                                Subtype_Is_Static,
   --#                                Type_Sym,
   --#                                Upper &
   --#         SLI.State         from *,
   --#                                Accuracy,
   --#                                CommandLineData.Content,
   --#                                ContextManager.Ops.Unit_Heap,
   --#                                ContextManager.Ops.Unit_Stack,
   --#                                Dictionary.Dict,
   --#                                ErrorHandler.Error_Context,
   --#                                Ident_Node,
   --#                                Id_Str,
   --#                                LexTokenManager.State,
   --#                                Lower,
   --#                                Scope,
   --#                                STree.Table,
   --#                                Subtype_Is_Static,
   --#                                Type_Sym,
   --#                                Upper &
   --#         SPARK_IO.File_Sys from *,
   --#                                Accuracy,
   --#                                CommandLineData.Content,
   --#                                ContextManager.Ops.File_Heap,
   --#                                ContextManager.Ops.Unit_Heap,
   --#                                ContextManager.Ops.Unit_Stack,
   --#                                Dictionary.Dict,
   --#                                ErrorHandler.Error_Context,
   --#                                Ident_Node,
   --#                                Id_Str,
   --#                                LexTokenManager.State,
   --#                                Lower,
   --#                                Scope,
   --#                                SLI.State,
   --#                                STree.Table,
   --#                                Subtype_Is_Static,
   --#                                Type_Node_Pos,
   --#                                Type_Sym,
   --#                                Upper;
   is
      Subtype_Symbol : Dictionary.Symbol;
   begin
      -- ASSUME Ident_Node = identifier
      SystemErrors.RT_Assert
        (C       => Syntax_Node_Type (Node => Ident_Node) = SPSymbols.identifier,
         Sys_Err => SystemErrors.Invalid_Syntax_Tree,
         Msg     => "Expect Ident_Node = identifier in Add_Scalar_Subtype");

      if Dictionary.TypeIsInteger (Type_Sym) then
         Dictionary.AddIntegerSubtype
           (Name            => Id_Str,
            Static          => Subtype_Is_Static,
            Parent          => Type_Sym,
            ParentReference => Dictionary.Location'(Start_Position => Type_Node_Pos,
                                                    End_Position   => Type_Node_Pos),
            Lower           => Lower,
            Upper           => Upper,
            Comp_Unit       => ContextManager.Ops.Current_Unit,
            Declaration     => Dictionary.Location'(Start_Position => Node_Position (Node => Ident_Node),
                                                    End_Position   => Node_Position (Node => Ident_Node)),
            Scope           => Scope,
            Context         => Dictionary.ProgramContext,
            Subtype_Symbol  => Subtype_Symbol);
         if ErrorHandler.Generate_SLI then
            SLI.Generate_Xref_Symbol
              (Comp_Unit      => ContextManager.Ops.Current_Unit,
               Parse_Tree     => Ident_Node,
               Symbol         => Subtype_Symbol,
               Is_Declaration => True);
         end if;
      elsif Dictionary.TypeIsModular (Type_Sym) then
         Dictionary.AddModularSubtype
           (Name            => Id_Str,
            Parent          => Type_Sym,
            ParentReference => Dictionary.Location'(Start_Position => Type_Node_Pos,
                                                    End_Position   => Type_Node_Pos),
            Lower           => Lower,
            Upper           => Upper,
            Comp_Unit       => ContextManager.Ops.Current_Unit,
            Declaration     => Dictionary.Location'(Start_Position => Node_Position (Node => Ident_Node),
                                                    End_Position   => Node_Position (Node => Ident_Node)),
            Scope           => Scope,
            Context         => Dictionary.ProgramContext,
            Subtype_Symbol  => Subtype_Symbol);
         if ErrorHandler.Generate_SLI then
            SLI.Generate_Xref_Symbol
              (Comp_Unit      => ContextManager.Ops.Current_Unit,
               Parse_Tree     => Ident_Node,
               Symbol         => Subtype_Symbol,
               Is_Declaration => True);
         end if;
      elsif Dictionary.TypeIsEnumeration (Type_Sym) then
         Dictionary.AddEnumerationSubtype
           (Name            => Id_Str,
            Static          => Subtype_Is_Static,
            Parent          => Type_Sym,
            ParentReference => Dictionary.Location'(Start_Position => Type_Node_Pos,
                                                    End_Position   => Type_Node_Pos),
            Lower           => Lower,
            Upper           => Upper,
            Comp_Unit       => ContextManager.Ops.Current_Unit,
            Declaration     => Dictionary.Location'(Start_Position => Node_Position (Node => Ident_Node),
                                                    End_Position   => Node_Position (Node => Ident_Node)),
            Scope           => Scope,
            Context         => Dictionary.ProgramContext,
            Subtype_Symbol  => Subtype_Symbol);
         if ErrorHandler.Generate_SLI then
            SLI.Generate_Xref_Symbol
              (Comp_Unit      => ContextManager.Ops.Current_Unit,
               Parse_Tree     => Ident_Node,
               Symbol         => Subtype_Symbol,
               Is_Declaration => True);
         end if;
      elsif Dictionary.TypeIsFloatingPoint (Type_Sym) then
         Dictionary.AddFloatingPointSubtype
           (Name            => Id_Str,
            Static          => Subtype_Is_Static,
            Parent          => Type_Sym,
            ParentReference => Dictionary.Location'(Start_Position => Type_Node_Pos,
                                                    End_Position   => Type_Node_Pos),
            Lower           => Lower,
            Upper           => Upper,
            ErrorBound      => Accuracy,
            Comp_Unit       => ContextManager.Ops.Current_Unit,
            Declaration     => Dictionary.Location'(Start_Position => Node_Position (Node => Ident_Node),
                                                    End_Position   => Node_Position (Node => Ident_Node)),
            Scope           => Scope,
            Context         => Dictionary.ProgramContext,
            Subtype_Symbol  => Subtype_Symbol);
         if ErrorHandler.Generate_SLI then
            SLI.Generate_Xref_Symbol
              (Comp_Unit      => ContextManager.Ops.Current_Unit,
               Parse_Tree     => Ident_Node,
               Symbol         => Subtype_Symbol,
               Is_Declaration => True);
         end if;
      elsif Dictionary.TypeIsFixedPoint (Type_Sym) then
         Dictionary.AddFixedPointSubtype
           (Name            => Id_Str,
            Static          => Subtype_Is_Static,
            Parent          => Type_Sym,
            ParentReference => Dictionary.Location'(Start_Position => Type_Node_Pos,
                                                    End_Position   => Type_Node_Pos),
            Lower           => Lower,
            Upper           => Upper,
            ErrorBound      => Accuracy,
            Comp_Unit       => ContextManager.Ops.Current_Unit,
            Declaration     => Dictionary.Location'(Start_Position => Node_Position (Node => Ident_Node),
                                                    End_Position   => Node_Position (Node => Ident_Node)),
            Scope           => Scope,
            Context         => Dictionary.ProgramContext,
            Subtype_Symbol  => Subtype_Symbol);
         if ErrorHandler.Generate_SLI then
            SLI.Generate_Xref_Symbol
              (Comp_Unit      => ContextManager.Ops.Current_Unit,
               Parse_Tree     => Ident_Node,
               Symbol         => Subtype_Symbol,
               Is_Declaration => True);
         end if;
      end if;
   end Add_Scalar_Subtype;

   -----------------------------------------------------------------------

   function Already_Defined (Ident_Str : LexTokenManager.Lex_String;
                             Scope     : Dictionary.Scopes) return Boolean
   --# global in Dictionary.Dict;
   --#        in LexTokenManager.State;
   is
      Sym : Dictionary.Symbol;
   begin
      Sym := Dictionary.LookupItem (Name              => Ident_Str,
                                    Scope             => Scope,
                                    Context           => Dictionary.ProofContext,
                                    Full_Package_Name => False);

      return not (Sym = Dictionary.NullSymbol
                    or else (Dictionary.IsTypeMark (Sym)
                               and then Dictionary.TypeIsAnnounced (Sym)
                               and then not Dictionary.IsDeclared (Sym)));
   end Already_Defined;

   -----------------------------------------------------------------------

   function Select_Index_Or_Discriminant_Error (Constraint_Node : STree.SyntaxNode) return Natural
   --# global in CommandLineData.Content;
   --#        in STree.Table;
   is
      Result : Natural;
   begin
      -- ASSUME Constraint_Node = index_or_discriminant_constraint
      SystemErrors.RT_Assert
        (C       => Syntax_Node_Type (Node => Constraint_Node) = SPSymbols.index_or_discriminant_constraint,
         Sys_Err => SystemErrors.Invalid_Syntax_Tree,
         Msg     => "Expect Constraint_Node = index_or_discriminant_constraint in Select_Index_Or_Discriminant_Error");

      -- if an index_or_discriminant_constraint is applied to an inapplicable type we need to report and
      -- error.  This function tries to narrow the scope of the error message returned.
      if CommandLineData.Ravenscar_Selected then
         -- we could be expecting an array, task or protected type
         if Is_Named_Association (Node => Constraint_Node) then
            -- must be Task or protected
            Result := 891;
         else
            -- could be any of Task, Protected, Array
            Result := 892;
         end if;
      else
         -- can only be an array
         Result := 41;
      end if;
      return Result;
   end Select_Index_Or_Discriminant_Error;

   -----------------------------------------------------------------------

   procedure Wf_Ravenscar_Subtype
     (Id_Str          : in LexTokenManager.Lex_String;
      Type_Sym        : in Dictionary.Symbol;
      Scope           : in Dictionary.Scopes;
      Id_Node         : in STree.SyntaxNode;
      Constraint_Node : in STree.SyntaxNode)
   --# global in     CommandLineData.Content;
   --#        in     ContextManager.Ops.File_Heap;
   --#        in     ContextManager.Ops.Unit_Heap;
   --#        in     ContextManager.Ops.Unit_Stack;
   --#        in out AggregateStack.State;
   --#        in out Dictionary.Dict;
   --#        in out ErrorHandler.Error_Context;
   --#        in out LexTokenManager.State;
   --#        in out SLI.State;
   --#        in out SPARK_IO.File_Sys;
   --#        in out Statistics.TableUsage;
   --#        in out STree.Table;
   --#        in out TheHeap;
   --# derives AggregateStack.State,
   --#         Dictionary.Dict,
   --#         LexTokenManager.State,
   --#         Statistics.TableUsage,
   --#         STree.Table,
   --#         TheHeap                    from *,
   --#                                         CommandLineData.Content,
   --#                                         Constraint_Node,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         Id_Node,
   --#                                         Id_Str,
   --#                                         LexTokenManager.State,
   --#                                         Scope,
   --#                                         STree.Table,
   --#                                         Type_Sym &
   --#         ErrorHandler.Error_Context,
   --#         SLI.State,
   --#         SPARK_IO.File_Sys          from CommandLineData.Content,
   --#                                         Constraint_Node,
   --#                                         ContextManager.Ops.File_Heap,
   --#                                         ContextManager.Ops.Unit_Heap,
   --#                                         ContextManager.Ops.Unit_Stack,
   --#                                         Dictionary.Dict,
   --#                                         ErrorHandler.Error_Context,
   --#                                         Id_Node,
   --#                                         Id_Str,
   --#                                         LexTokenManager.State,
   --#                                         Scope,
   --#                                         SLI.State,
   --#                                         SPARK_IO.File_Sys,
   --#                                         STree.Table,
   --#                                         Type_Sym;
      is separate;

begin -- Wf_Subtype_Declaration

   -- ASSUME Node = subtype_declaration
   SystemErrors.RT_Assert
     (C       => Syntax_Node_Type (Node => Node) = SPSymbols.subtype_declaration,
      Sys_Err => SystemErrors.Invalid_Syntax_Tree,
      Msg     => "Expect Node = subtype_declaration in Wf_Subtype_Declaration");

   Lower    := LexTokenManager.Null_String;
   Upper    := LexTokenManager.Null_String;
   Accuracy := LexTokenManager.Null_String;

   Ident_Node := Child_Node (Current_Node => Node);
   -- ASSUME Ident_Node = identifier
   SystemErrors.RT_Assert
     (C       => Syntax_Node_Type (Node => Ident_Node) = SPSymbols.identifier,
      Sys_Err => SystemErrors.Invalid_Syntax_Tree,
      Msg     => "Expect Ident_Node = identifier in Wf_Subtype_Declaration");
   Id_Str := Node_Lex_String (Node => Ident_Node);

   if Already_Defined (Ident_Str => Id_Str,
                       Scope     => Scope) then
      Ok_To_Add := False;
      ErrorHandler.Semantic_Error
        (Err_Num   => 10,
         Reference => ErrorHandler.No_Reference,
         Position  => Node_Position (Node => Ident_Node),
         Id_Str    => Id_Str);
   else
      Ok_To_Add := True;
   end if;

   Type_Node := Child_Node (Current_Node => Next_Sibling (Current_Node => Ident_Node));
   -- ASSUME Type_Node = type_mark
   SystemErrors.RT_Assert
     (C       => Syntax_Node_Type (Node => Type_Node) = SPSymbols.type_mark,
      Sys_Err => SystemErrors.Invalid_Syntax_Tree,
      Msg     => "Expect Type_Node = type_mark in Wf_Subtype_Declaration");

   Wf_Type_Mark (Node          => Type_Node,
                 Current_Scope => Scope,
                 Context       => Dictionary.ProgramContext,
                 Type_Sym      => Type_Sym);

   if Dictionary.IsPrivateType (Type_Sym, Scope) then
      Ok_To_Add := False;
      ErrorHandler.Semantic_Error
        (Err_Num   => 48,
         Reference => ErrorHandler.No_Reference,
         Position  => Node_Position (Node => Type_Node),
         Id_Str    => LexTokenManager.Null_String);
   end if;

   -- Subtypes of generic types are not allowed because we can't check whether the boudns will be valid
   -- when they are instantiated
   if Dictionary.TypeIsGeneric (Type_Sym) then
      Ok_To_Add := False;
      ErrorHandler.Semantic_Error
        (Err_Num   => 652,
         Reference => ErrorHandler.No_Reference,
         Position  => Node_Position (Node => Type_Node),
         Id_Str    => LexTokenManager.Null_String);
   end if;

   Constraint_Node := Next_Sibling (Current_Node => Type_Node);
   -- ASSUME Constraint_Node = constraint OR NULL
   if Syntax_Node_Type (Node => Constraint_Node) = SPSymbols.constraint then
      Constraint_Found := True;
   elsif Constraint_Node = STree.NullNode then
      Constraint_Found := False;
   else
      Constraint_Found := False;
      SystemErrors.Fatal_Error
        (Sys_Err => SystemErrors.Invalid_Syntax_Tree,
         Msg     => "Expect Constraint_Node = constraint OR NULL in Wf_Subtype_Declaration");
   end if;

   -- test to prevent Boolean subtype unless full-range
   if Constraint_Found and then Dictionary.TypeIsBoolean (Type_Sym) then
      Ok_To_Add := False;
      ErrorHandler.Semantic_Error
        (Err_Num   => 412,
         Reference => 15,
         Position  => Node_Position (Node => Constraint_Node),
         Id_Str    => LexTokenManager.Null_String);
   end if;

   if Ok_To_Add and then not Dictionary.IsUnknownTypeMark (Type_Sym) then
      if Constraint_Found then
         Constraint_Node := Child_Node (Current_Node => Constraint_Node);
         -- ASSUME Constraint_Node = range_constraint OR floating_point_constraint OR fixed_point_constraint OR
         --                          index_or_discriminant_constraint
         -- there is a constraint node so proceed as before
         if Syntax_Node_Type (Node => Constraint_Node) = SPSymbols.index_or_discriminant_constraint then
            -- ASSUME Constraint_Node = index_or_discriminant_constraint
            if Dictionary.TypeIsArray (Type_Sym) then
               if Is_Named_Association (Node => Constraint_Node) then
                  ErrorHandler.Semantic_Error
                    (Err_Num   => 92,
                     Reference => ErrorHandler.No_Reference,
                     Position  => Node_Position (Node => Constraint_Node),
                     Id_Str    => LexTokenManager.Null_String);
               else -- positional association is ok
                  if Dictionary.IsUnconstrainedArrayType (Type_Sym) then
                     Dictionary.AddArraySubtype
                       (Name            => Id_Str,
                        Parent          => Type_Sym,
                        ParentReference => Dictionary.Location'(Start_Position => Node_Position (Node => Type_Node),
                                                                End_Position   => Node_Position (Node => Type_Node)),
                        Comp_Unit       => ContextManager.Ops.Current_Unit,
                        Declaration     => Dictionary.Location'(Start_Position => Node_Position (Node => Ident_Node),
                                                                End_Position   => Node_Position (Node => Ident_Node)),
                        Scope           => Scope,
                        Context         => Dictionary.ProgramContext,
                        Static          => False,
                        ArraySubtype    => Subtype_Sym);
                     if ErrorHandler.Generate_SLI then
                        SLI.Generate_Xref_Symbol
                          (Comp_Unit      => ContextManager.Ops.Current_Unit,
                           Parse_Tree     => Ident_Node,
                           Symbol         => Subtype_Sym,
                           Is_Declaration => True);
                     end if;
                     Check_Index_Constraints
                       (Constraint_Node => Constraint_Node,
                        Scope           => Scope,
                        Subtype_Sym     => Subtype_Sym,
                        Type_Sym        => Type_Sym,
                        All_Indexes_OK  => All_Indexes_OK);
                     Dictionary.SetTypeIsWellformed (Subtype_Sym, All_Indexes_OK);
                  else -- array already constrained
                     ErrorHandler.Semantic_Error
                       (Err_Num   => 99,
                        Reference => ErrorHandler.No_Reference,
                        Position  => Node_Position (Node => Type_Node),
                        Id_Str    => LexTokenManager.Null_String);
                  end if;
               end if; -- illegal named association fo index_constraint
            elsif Dictionary.IsProtectedType (Type_Sym) or else Dictionary.IsTaskType (Type_Sym) then
               Wf_Ravenscar_Subtype
                 (Id_Str          => Id_Str,
                  Type_Sym        => Type_Sym,
                  Scope           => Scope,
                  Id_Node         => Ident_Node,
                  Constraint_Node => Constraint_Node);
            else
               -- a type has been supplied for which index_or_discriminant_constraint is
               -- not appropriate
               ErrorHandler.Semantic_Error
                 (Err_Num   => Select_Index_Or_Discriminant_Error (Constraint_Node => Constraint_Node),
                  Reference => ErrorHandler.No_Reference,
                  Position  => Node_Position (Node => Type_Node),
                  Id_Str    => LexTokenManager.Null_String);
            end if;
         else
            -- ASSUME Constraint_Node = range_constraint OR floating_point_constraint OR fixed_point_constraint
            -- some scalar subtype expected
            if not Dictionary.TypeIsScalar (Type_Sym) then
               ErrorHandler.Semantic_Error
                 (Err_Num   => 59,
                  Reference => ErrorHandler.No_Reference,
                  Position  => Node_Position (Node => Type_Node),
                  Id_Str    => LexTokenManager.Null_String);
            else
               Subtype_Is_Static := True; -- default value
               if Syntax_Node_Type (Node => Constraint_Node) = SPSymbols.range_constraint then
                  -- ASSUME Constraint_Node = range_constraint
                  Check_Range
                    (Node     => Child_Node (Current_Node => Constraint_Node),
                     Type_Sym => Type_Sym,
                     Scope    => Scope,
                     Static   => Subtype_Is_Static,
                     Lower    => Lower,
                     Upper    => Upper);

                  -- if constraint is a range but type is real then no accuracy
                  -- has been supplied so we need to get it from parent
                  if Dictionary.TypeIsFloatingPoint (Type_Sym) then
                     Accuracy := Dictionary.GetScalarAttributeValue (False, LexTokenManager.Digits_Token, Type_Sym);
                  elsif Dictionary.TypeIsFixedPoint (Type_Sym) then
                     Accuracy := Dictionary.GetScalarAttributeValue (False, LexTokenManager.Delta_Token, Type_Sym);
                  end if;
                  Subtype_Is_Static := Subtype_Is_Static and then Dictionary.IsStatic (Type_Sym, Scope);

               elsif Syntax_Node_Type (Node => Constraint_Node) = SPSymbols.floating_point_constraint then
                  -- ASSUME Constraint_Node = floating_point_constraint
                  if not Dictionary.TypeIsFloatingPoint (Type_Sym) then
                     ErrorHandler.Semantic_Error
                       (Err_Num   => 100,
                        Reference => ErrorHandler.No_Reference,
                        Position  => Node_Position (Node => Type_Node),
                        Id_Str    => LexTokenManager.Null_String);
                  else
                     Check_Real_Accuracy
                       (Node     => Constraint_Node,
                        Scope    => Scope,
                        Sort     => Is_Floating,
                        Accuracy => Accuracy,
                        Static   => Subtype_Is_Static);
                     Check_Real_Range
                       (Node     => Constraint_Node,
                        Type_Sym => Type_Sym,
                        Scope    => Scope,
                        Static   => Subtype_Is_Static,
                        Lower    => Lower,
                        Upper    => Upper);
                  end if;

               elsif Syntax_Node_Type (Node => Constraint_Node) = SPSymbols.fixed_point_constraint then
                  -- ASSUME Constraint_Node = fixed_point_constraint
                  if not Dictionary.TypeIsFixedPoint (Type_Sym) then
                     ErrorHandler.Semantic_Error
                       (Err_Num   => 101,
                        Reference => ErrorHandler.No_Reference,
                        Position  => Node_Position (Node => Type_Node),
                        Id_Str    => LexTokenManager.Null_String);
                  else
                     Check_Real_Accuracy
                       (Node     => Constraint_Node,
                        Scope    => Scope,
                        Sort     => Is_Fixed,
                        Accuracy => Accuracy,
                        Static   => Subtype_Is_Static);
                     Check_Real_Range
                       (Node     => Constraint_Node,
                        Type_Sym => Type_Sym,
                        Scope    => Scope,
                        Static   => Subtype_Is_Static,
                        Lower    => Lower,
                        Upper    => Upper);
                  end if;
               else
                  SystemErrors.Fatal_Error
                    (Sys_Err => SystemErrors.Invalid_Syntax_Tree,
                     Msg     => "Expect Constraint_Node = range_constraint OR floating_point_constraint OR fixed_point_constraint OR index_or_discriminant_constraint in Wf_Subtype_Declaration");
               end if;
               Add_Scalar_Subtype
                 (Accuracy          => Accuracy,
                  Id_Str            => Id_Str,
                  Ident_Node        => Ident_Node,
                  Type_Node_Pos     => Node_Position (Node => Type_Node),
                  Type_Sym          => Type_Sym,
                  Scope             => Scope,
                  Subtype_Is_Static => Subtype_Is_Static,
                  Lower             => Lower,
                  Upper             => Upper);
            end if;
         end if;
      else  -- no constraint node present
         if Dictionary.TypeIsScalar (Type_Sym) then
            --scalars are allowed without constraints so process them here
            Subtype_Is_Static := Dictionary.IsStatic (Type_Sym, Scope);
            Lower             := Dictionary.GetScalarAttributeValue (False, LexTokenManager.First_Token, Type_Sym);
            Upper             := Dictionary.GetScalarAttributeValue (False, LexTokenManager.Last_Token, Type_Sym);
            if Dictionary.TypeIsFloatingPoint (Type_Sym) then
               Accuracy := Dictionary.GetScalarAttributeValue (False, LexTokenManager.Digits_Token, Type_Sym);
            elsif Dictionary.TypeIsFixedPoint (Type_Sym) then
               Accuracy := Dictionary.GetScalarAttributeValue (False, LexTokenManager.Delta_Token, Type_Sym);
            end if;
            Add_Scalar_Subtype
              (Accuracy          => Accuracy,
               Id_Str            => Id_Str,
               Ident_Node        => Ident_Node,
               Type_Node_Pos     => Node_Position (Node => Type_Node),
               Type_Sym          => Type_Sym,
               Scope             => Scope,
               Subtype_Is_Static => Subtype_Is_Static,
               Lower             => Lower,
               Upper             => Upper);
            -- Full-range subtypes are also allowed for non-tagged records.
         elsif Dictionary.TypeIsRecord (Type_Sym) and then not Dictionary.TypeIsTagged (Type_Sym) then
            Dictionary.AddRecordSubtype
              (Name            => Id_Str,
               Parent          => Type_Sym,
               ParentReference => Dictionary.Location'(Start_Position => Node_Position (Node => Type_Node),
                                                       End_Position   => Node_Position (Node => Type_Node)),
               Comp_Unit       => ContextManager.Ops.Current_Unit,
               Declaration     => Dictionary.Location'(Start_Position => Node_Position (Node => Ident_Node),
                                                       End_Position   => Node_Position (Node => Ident_Node)),
               Scope           => Scope,
               Context         => Dictionary.ProgramContext,
               Subtype_Symbol  => Subtype_Symbol);
            if ErrorHandler.Generate_SLI then
               SLI.Generate_Xref_Symbol
                 (Comp_Unit      => ContextManager.Ops.Current_Unit,
                  Parse_Tree     => Ident_Node,
                  Symbol         => Subtype_Symbol,
                  Is_Declaration => True);
            end if;
         elsif Dictionary.TypeIsArray (Type_Sym) and then not Dictionary.IsUnconstrainedArrayType (Type_Sym) then

            Dictionary.AddFullRangeArraySubtype
              (Name            => Id_Str,
               Parent          => Type_Sym,
               ParentReference => Dictionary.Location'(Start_Position => Node_Position (Node => Type_Node),
                                                       End_Position   => Node_Position (Node => Type_Node)),
               Comp_Unit       => ContextManager.Ops.Current_Unit,
               Declaration     => Dictionary.Location'(Start_Position => Node_Position (Node => Ident_Node),
                                                       End_Position   => Node_Position (Node => Ident_Node)),
               Scope           => Scope,
               Context         => Dictionary.ProgramContext,
               Subtype_Symbol  => Subtype_Symbol);
            if ErrorHandler.Generate_SLI then
               SLI.Generate_Xref_Symbol
                 (Comp_Unit      => ContextManager.Ops.Current_Unit,
                  Parse_Tree     => Ident_Node,
                  Symbol         => Subtype_Symbol,
                  Is_Declaration => True);
            end if;

         else -- not scalar or non-tagged record so must be illegal
            ErrorHandler.Semantic_Error
              (Err_Num   => 406,
               Reference => ErrorHandler.No_Reference,
               Position  => Node_Position (Node => Type_Node),
               Id_Str    => LexTokenManager.Null_String);
         end if;
      end if;
   end if;
end Wf_Subtype_Declaration;
